/*
 * Projekt: C-quest
 * Autor: David Martinek (c) 2004
 * email: martinek@fit.vutbr.cz
 * Licence: GPL v.2
 * Popis: Modul pro spoutn programu v testovacm a vukovm mdu.
 *
 * !
 * ! Tento modul pouv mrn pokroilej programtorsk techniky, kter
 * ! jdou za rmec kurzu programovn pro zatenky. Zejmna jde o podmnn
 * ! peklad, pouvn nestandardnch knihoven a deklarace funkc s promnnm
 * ! potem parametr. 
 * !
 *  
 * $Id: cqlearntest.c,v 1.9 2004/07/06 15:20:23 martinek Exp $
 * 
 * Copyright (C)2004 David Martinek
 *
 * Tento program je svobodn software; mete jej it a modifikovat podle
 * ustanoven GNU General Public License, vydvan Free Software
 * Foundation; a to bu verze 2 tto licence anebo (podle vaeho uven)
 * kterkoli pozdj verze.
 *
 * Tento program je roziovn v nadji, e bude uiten, avak BEZ
 * JAKKOLI ZRUKY; neposkytuj se ani odvozen zruky PRODEJNOSTI anebo
 * VHODNOSTI PRO URIT EL. Dal podrobnosti hledejte ve GNU General Public 
 * License.
 *
 * Kopii GNU General Public License jste ml obdret spolu s tmto
 * programem; pokud se tak nestalo, napite o ni Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "cqlearntest.h"
#include "cqtools.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>


// podmnn peklad
#ifdef YESCURSES
#include <curses.h>
#endif

/**
 * Tato funkce slou jako obal pro funkce printf a printw. Pokud je definovno
 * makro YESCURSES, pouije se funkce printw, jinak printf.
 */
void print(const char *format, ...)
{
  va_list params;
  va_start(params, format);
#ifdef YESCURSES 
  vwprintw(stdscr, format, params);
#else
  vprintf(format, params);
#endif  
  va_end(params);
}

/**
 * Konstanta pouvan funkc showQuestion, jako volba pro vytisknut vech
 * monch odpovd.
 */
const char CHOICE_ALL = '\0';

/**
 * Zobraz otzku a nabdne sprvn odpovdi.
 * @param item Ukazatel na poloku seznamu otzek.
 * @param choice Udv odpov, kterou m zobrazit. Pokud m hodnotu 
 *   CHOICE_ALL, zobraz vechny odpovdi.
 */
void showQuestion(TItem *item, char choice)
{
  assert(item != NULL);
  // tohle by mlo bt oeteno u pi natn souboru
  assert(choice == CHOICE_ALL || (choice >= 'a' && choice <= 'z'));
  
  print("Otzka:\n");
  print("%s", item->questLines);
  print("-----------------------------------------------------------------"
         "------------\n");
  if (choice == CHOICE_ALL)
  {
    print("Monosti:\n");
    print("%s", item->answLines);
  }
  else
  {
    print("Sprvn odpov:\n");

    char tmp[strlen(item->answLines+1)];
    strcpy(tmp, item->answLines);
    char *answer = strtok(tmp, "\n");
    while (answer != NULL)
    {
      if (answer[0] == choice)
      {
        //adresovou aritmetikou peskakuju psmeno na zatku
        print("%s\n", answer+1);
        break;
      }
      answer = strtok(NULL, "\n");
    }
    assert(answer != NULL); // nenae poadovanou odpov -> chyba
  }
  print("-----------------------------------------------------------------"
         "------------\n");
}

/**
 * Vype na curses obrazovku vysvtlen ke sprvn odpovdi.
 */
void showExplanation(TItem *item)
{
  assert(item != NULL);
  
  print("Zdvodnn:\n");
  print("%s", item->explLines);
  print("\n");
}

/**
 * Pehod hodnoty dvou ukazatel.
 */
static inline void swapItems(TItem *a, TItem *b)
{
  assert(a != NULL && b != NULL);

  TItem tmp = *a;
  *a = *b;
  *b = tmp;  
}

/**
 * Vrt nhodn slo od 0 do max-1.
 */
static inline unsigned int getRandom(unsigned int max)
{
  return (unsigned int)rand() % max;
}

/**
 * Nhodn zamch pole ukazatel na poloky testu.
 */
void shuffle(TItem *items[], unsigned int length)
{
  for (unsigned int i = 0; i < length; i++)
  {
    unsigned int j = getRandom(length);
    swapItems(items[i], items[j]);
  }
}

/**
 * Zkopruje ukazatele na prvky seznamu do pole a pak jej nhodn zamch. Pole
 * items mus mt alokovno minimln data->total poloek.
 */
void copyAndShuffle(TItem *items[], const TTestData *data)
{
  assert(data != NULL);

  srand(time(NULL));  // inicializace genertoru nhodnch sel

  TItem *tmp = data->first;
  for (unsigned int i = 0; i < data->total; i++)
  {
    items[i] = tmp;
    tmp = tmp->next;
  }
  shuffle(items, data->total); // zamch otzky
}

/**
 * Vytiskne zprvu msg a pok, dokud uivatel nestiskne klvesu, kterou potom
 * vrt jako vsledek. Tato funkce pracuje s curses.
 */
static inline int waitForKey(char *msg)
{
  print(msg);
#ifdef YESCURSES  
  int c = getch();
  if (c == '\n' || c == EOF) return c;
  while (getch() != '\n')
    ; 
#else
  int c = getchar();
  if (c == '\n' || c == EOF) return c;
  while (getchar() != '\n')
    ;
#endif
  return c;
}

/**
 * Vype na obrazovku vsledky. Vypisuje na stdout, ne na curses obrazovku.
 * @param title    Nzev testu.
 * @param right    Poet sprvn zodpovzench otzek.
 * @param answered Poet zodpovzench otzek.
 * @param total    Celkov poet otzek.
 */
void printResult(char *title, int right, int answered, int total)
{
  printf("*****************************************************************"
         "************\n");
  printf("%s", title);
  printf("Zodpovzeno %d otzek z %d celkem.\n", answered, total);
  printf("V vsledek:\n");
  printf("  Sprvn:              %d odpovd z %d otzek (%d%%)\n", 
      right, total, (right*100)/total);
  printf("  patn/nezodpovzeno: %d odpovd z %d otzek (%d%%)\n",
      total-right, total, 100 - (right*100)/total);
  printf("*****************************************************************"
         "************\n");
}

/**
 * Vytiskne, kolik jet zbv otzek, podle esk gramatiky.
 */
void showRemain(unsigned int remain)
{
  if (remain == 1)
  {
    print("Zbv 1 otzka.\n");
  }
  else if (remain >= 2 && remain <= 4)
  {
    print("Zbvaj %d otzky.\n", remain);
  }
  else
  {
    print("Zbv %d otzek.\n", remain);
  }
}

/**
 * Vymae obrazovku.
 */
void myErase()
{
#ifdef YESCURSES
  erase();
#else
  for (int i = 0; i < 50; i++)
  {
    printf("\n");
  }
#endif
}

/**
 * Inicializuje knihovnu ncurses.
 */
#ifdef YESCURSES  
void initCurses()
{
  initscr();  // start curses mdu
  nocbreak(); // getch bude ekat na enter
  echo();     // to, co getch nate, se tak zobraz na obrazovku  
}
#endif

/**
 * Spust program v testovacm mdu.
 *
 * Na zatku vechny testov otzky nhodn zamch. Pot postupn zobrazuje
 * otzky a ek na odpovdi. Po kad otzce zobraz text objasujc sprvnou
 * odpov. Pot poet sprvnch a patnch odpovd a na konci zobraz 
 * vsledek. 
 *
 * as se pot pouze pi ekn na odpov. Program me ekat na zobrazen
 * dal otzky libovoln dlouho, ani by se tento as potal do celkovho
 * asu vyhrazenho pro test hodnotou data->totalTime.
 *
 * Nemn datovou strukturu data.
 *
 * @param data Datov struktura obsahujc cel test.
 */
void runTest(const TTestData *data)
{
  assert(data != NULL);

  TItem *test[data->total];
  copyAndShuffle(test, data);

  long timeLeft = data->totalTime;
  int right = 0, wrong = 0;
  
#ifdef YESCURSES  
  initCurses();
#endif
  
  for (unsigned int i = 0; i < data->total; i++)
  {
    myErase();
    print("*************************** Zbv %6d sekund ********************"
        "*********\n", timeLeft);
    showQuestion(test[i], CHOICE_ALL);
    time_t t1 = time(NULL);
    
    if (waitForKey("Vae odpov: ") == test[i]->rightAnswer)
    {
      print("\n=== Sprvn ===\n");
      right += 1;
    }
    else
    {
      print("\nXXX patn. Sprvn odpov je %c. XXX\n", 
          test[i]->rightAnswer);
      wrong += 1;
    }
    
    time_t t2 = time(NULL);
    timeLeft -= t2 - t1;
    
    print("-----------------------------------------------------------------"
        "------------\n");   
    showExplanation(test[i]);
    
    if (timeLeft <= 0)
    {
      waitForKey("Vyprel as, vyhodnocen se zobraz po stisku klvesy "
          "Enter.\n");
      break;
    }
    
    if (i < data->total-1)
    {
      showRemain(data->total - i - 1);
      int c = waitForKey("Pro ukonen napite q, pro pokraovn stisknte"
          " Enter.\n");
      if (c == 'q' || c == 'Q') break;
    }
    else
    {
      waitForKey("Vyhodnocen se zobrac po stisku klvesy Enter.\n");
    }
  }

#ifdef YESCURSES  
  endwin(); // konec curses mdu
#endif
  printResult(data->title, right, right+wrong, data->total);
}

/**
 * Spust program v mdu uen. V tomto mdu se nepot ani as, ani skre.
 * Program neek na odpovdi od uivatele, ale sm je rovnou zobraz spolu
 * s vysvtlenm.
 *
 * @param data Datov struktura obsahujc cel test.
 */
void runLearning(const TTestData *data)
{
  assert(data != NULL);

  TItem *test[data->total];
  copyAndShuffle(test, data);

#ifdef YESCURSES  
  initCurses(); // start curses mdu
#endif  
  
  for (unsigned int i = 0; i < data->total; i++)
  {
    myErase();

    showQuestion(test[i], test[i]->rightAnswer);
    showExplanation(test[i]);
    
    if (i < data->total-1)
    {
      showRemain(data->total - i);
      int c = waitForKey("Pro ukonen napite q, pro pokraovn stisknte"
          " Enter.\n");
      if (c == 'q' || c == 'Q') break;
    }
    else
    {
      waitForKey("Vechny otzky byly vyerpny, program skon po stisku "
                 "klvesy Enter.\n");
    }
  }

#ifdef YESCURSES
  endwin(); // konec curses mdu
#endif

  printf("%s", data->title);
  printf("Konec vukovho mdu.\n");
}

